#############################################################################
::xotcl::THREAD create throttleThread {

  Class ThrottleStat -parameter { type user_id timestamp ip_adress url }

  Class Throttle -parameter {
    {timeWindow 10}
    {timeoutMs 2000}
    {startThrottle 7} 
    {toMuch 10} 
    {alerts 0} {throttles 0} {rejects 0} {last_mkey ""} 
  }

  Throttle instproc init {} {
    Object create [self]::stats
    Object create [self]::users
    next
  }

  Throttle instproc add_stat { type user_id ip_adress url query } {
    set furl [expr {$query != "" ? "$url?$query" : $url}]
    set entry [ThrottleStat new -childof [self]::stats \
		   -type $type -user_id $user_id -timestamp [clock seconds] \
		   -ip_adress $ip_adress -url $furl]
  }

  Throttle instproc url_statistics {{flush 0}} {
    set data [lsort -dictionary [[self]::stats info children]]
    if { [llength $data] == 0} {
      return "No statistics available"
    } elseif {$flush} {
      foreach c $data {$c destroy}
      return ""
    } else {
      set output "<table><tr><th>Type</th><th>User ID</th><th>Time</th>
		<th>IP Adress</th><th>URL</th></tr>"
      foreach stat [lsort -dictionary [[self]::stats info children]] {
	append output "<tr><td>[$stat type]</td>" \
	    "<td>[$stat user_id]</a></td>" \
	    "<td>[clock format [$stat timestamp]]</td>" \
	    "<td>[$stat ip_adress]</td>" \
	    "<td>[$stat url]</td></tr>"
      }
      append output "</table>\n"
      return $output
    }
  }

  Throttle instproc call_statistics {} {
    set l [list]
    foreach t {seconds minutes hours} {
      lappend l [list $t [$t set last] [$t set trend] [$t set stats]]
    }
    return $l
  }

  Throttle instproc register_access {requestKey pa url} {
    my instvar last_mkey timeWindow
    set now 	[clock seconds]
    set mkey 	[expr { ($now / 60) % $timeWindow}]
    set obj 	[self]::users::$mkey

    if { $mkey ne $last_mkey} { 
      if {[string compare "" $last_mkey]} {my purge_access_stats $now}
      # create or recreate the container object for that minute
      if {[my isobject $obj]} {$obj destroy}
      Users create $obj -point_in_time $now
      my last_mkey $mkey
    } 
    $obj addKey $requestKey $pa $url
    Users expSmooth $now $requestKey
  }

  Throttle instproc purge_access_stats {now} {
    my instvar last_mkey timeWindow
    # purge stale entries (for low traffic)
    set secs [expr {$timeWindow*60}]
    if { $now - [[self]::users::$last_mkey point_in_time] > $secs } {
      # no requests for a while; delete all objects under [self]::users::
      Object create [self]::users
    } else {
      # delete selectively
      foreach element [[self]::users info children] {
        if { [$element point_in_time] < $now - $secs } {$element destroy}
      }
    } 
  }

  Throttle instproc running {} {
    my array get running_url
  }

  Throttle instproc check {requestKey pa url start_clicks content_type} {
    seconds ++
    set var running_url($requestKey,$url)
    # check first, whether the same user has already the same request
    # issued; if yes, block this request. Caveat: some html-pages
    # use the same image in many places, so we can't block it. This
    # will make the sttistics for images look better than they are.
    set is_image_request [string match image/* $content_type]
    if {[my exists $var] && !$is_image_request} {
      set ms [my set $var]
      my log "### already $var since [expr {[clock click -milliseconds]-$ms}]"
      return [list 0 0 1]
    } else {
      my set $var $start_clicks
      #my log "### new $var"
    }
    my register_access $requestKey $pa $url

    # ... the number of 14 is arbitrary. on our one real request
    # might have up to 14 subrequests (through iframes).... 
    if {$is_image_request || [my array size running_url] < 14} {

      # less than forteen running requests, let people do what they want;
      # don't throttle/block embedded images.....
      return [list 0 0 0]

    } else {

      # Check, whether the last request from a user was within 
      # the minimum time interval. We are not keeping a full table
      # of all request keys, but use an timeout triggered mechanism 
      # to keep only the active request keys in an associative array

      my incr alerts
      if {[my exists active($requestKey)]} {
	# if more than one request for this key is already active, return blocking time
	foreach {to ms cnt} [my set active($requestKey)] break
	set retMs [expr {$cnt>[my startThrottle] ? $start_clicks-$ms : 0}]
	# cancel the timeout
	after cancel $to
      } else {
	set retMs 0
	set cnt 0
      }
      incr cnt
      # establish a new timeout
      set to [after [my timeoutMs] [list [self] cancel $requestKey]]
      my set active($requestKey) [list $to $start_clicks $cnt]
      if {$cnt <= [my toMuch]} {
	set cnt 0
	if {$retMs>0} {my incr throttles}
      } else {
	my incr rejects
      }
      return [list $cnt $retMs 0]
    }
  }

  Throttle instproc statistics {} {
    return "<table>
        <tr><td>Number of alerts:</td><td>[my alerts]</td></tr>
        <tr><td>Number of throttles:</td><td>[my throttles]</td></tr>
        <tr><td>Number of rejects:</td><td>[my rejects]</td></tr>
        </table>\n"
  }

  Throttle instproc cancel {requestKey} {
    # cancel a timeout and clean up active request table for this key
    if {[my exists active($requestKey)]} {
      foreach {to ms} [my set active($requestKey)] break
      after cancel $to
      my unset active($requestKey)
      #my log "+++ Cancel $requestKey block"
    } else {
      my log "+++ Cancel for $requestKey failed !!!"
    }
  }

  Throttle instproc active { } {
    # return the currently active requests (for debugging and introspection)
    return [my array get active]
  }

  Throttle instproc add_url_stat {url time_used key} {
    catch {my unset running_url($key,$url)}
    #my log "### unset running_url($key,$url)"
    response_time_minutes add_url_stat $url $time_used
  }
  Throttle instproc report_url_stats {} {
    response_time_minutes report_url_stats
  }

  Throttle create throttle

  ############################
  # A simple counter class, which is able to aggregate values in some
  # higher level counters (report_to) and to keep statistics in form 
  # of a trend and max values)
  Class Counter -parameter { 
    report timeoutMs 
    {stats ""} {last ""} {trend ""} {c 0} {logging 0}
    {nr_trend_elements 48} {nr_stats_elements 5} 
  }
  Counter instproc ++ {} {
    my incr c
  }
  Counter instproc end {} {
    if {[my exists report]} {
      [my report] incr c [my c]
    }
    my finalize [my c]
    my c 0
  }
  Counter instproc log_to_file {timestamp label value} {
    if {![my logging]} return
    set server [ns_info server]
    set f [open $::logdir/counter.log a]
    puts $f "$timestamp -- $server $label $value"
    close $f
  } 
  Counter instproc finalize {n} {
    after cancel [my set to]
    my instvar stats trend
    #
    # trend keeps nr_trend_elements most recent values
    #
    lappend trend $n
    set lt [llength $trend]
    if {$lt > [my nr_trend_elements]} {
      set trend [lrange $trend [expr {$lt-[my nr_trend_elements]}] end]
    }
    #
    # stats keeps nr_stats_elements highest values with time stamp
    #
    set now [clock format [clock seconds]]
    lappend stats [list $now $n]
    set stats [lrange [lsort -real -decreasing -index 1 $stats] 0 [my nr_stats_elements]]
    #
    # log if necessary
    #
    catch {my log_to_file $now [self] $n}
    #
    my set to [after [my timeoutMs] [list [self] end]]
  }

  Counter instproc init {} {
    my set to [after [my timeoutMs] [list [self] end]]
    next
  }
  Counter instproc destroy {} {
    after cancel [my set to]
    next
  }

  Counter hours -timeoutMs [expr {60000*60}] -logging 1
  Counter minutes -timeoutMs 60000 -report hours -logging 1
  Counter seconds -timeoutMs 1000 -report minutes 

  Class MaxCounter -superclass Counter -instproc end {} {
    my c [Users total]
    if {[my exists report]} {
      [my report] instvar {c rc}
      if {$rc < [my c]} {set rc [my c]}
    }
    my finalize [my c]
    my c 0
  }

  MaxCounter user_count_hours -timeoutMs [expr {60000*60}] -logging 1
  MaxCounter user_count_minutes -timeoutMs 60000 -report user_count_hours -logging 1


  Class AvgCounter -superclass Counter -parameter {{t 0} {atleast 1}} -instproc end {} {
    if {[my c]>0} {
      set avg [expr {int([my t]*1.0/[my c])}]
    } else {
      set avg 0
    }
    if {[my exists report]} {
      [my report] incr c [my c]
      [my report] incr t [my t]
    }
    my finalize $avg
    my c 0
    my t 0
  }

  Class UrlCounter -superclass AvgCounter \
      -parameter {{truncate_check 10} {max_urls 0}} \
      -instproc add_url_stat {url ms} {
    my ++
    my incr t $ms
    if {[my exists stat($url)]} {
      my incr stat($url) $ms
      my incr cnt($url)
    } else {
      my set stat($url) $ms
      my set cnt($url) 1
    }
  } -instproc url_stats {} {
    set result [list]
    foreach url [my array names stat] {
      lappend result [list $url [my set stat($url)] [my set cnt($url)]]
    }
    set result [lsort -real -decreasing -index 1 $result]
    return $result
  } -instproc check_truncate_stats {} {
    # truncate statistics if necessary
    if {[info exists ::package_id]} {
      # get values from package parameters
      my max_urls [parameter::get -package_id $::package_id \
		       -parameter max-url-stats -default 13]
      # we use the timer to check other parameters as well here
      set time_window [parameter::get -package_id $::package_id \
			   -parameter time-window -default 10]
      if {$time_window != [throttle timeWindow]} {
	throttle timeWindow $time_window
	after 0 [list throttle purge_access_stats [clock seconds]]
      }
    }
    set max [my max_urls]
    if {$max>1} {
      set result [my url_stats]
      set l [llength $result]
      for {set i $max} {$i<$l} {incr i} {
	set url [lindex [lindex $result $i] 0]
	my unset stat($url)
	my unset cnt($url)
      }
      set result [lrange $result 0 [expr {$max-1}]]
      return $result
    }
    return ""
  } -instproc report_url_stats {} {
    set stats [my check_truncate_stats]
    if {[string equal "" $stats]} {
      set stats [my url_stats]
    }
    return $stats
  } -instproc finalize args {
    next
    # each time the timer runs out, perform the cleanup
    after 0 [list [self] check_truncate_stats]
  }


  UrlCounter response_time_hours -timeoutMs [expr {60000*60}] \
     -atleast 500 -logging 1
  UrlCounter response_time_minutes -timeoutMs 60000 \
     -report response_time_hours -atleast 100 \
     -logging 1

  #
  # Class for the user tracking

  Class Users -parameter { point_in_time } -ad_doc {
    This class is responsible for the user tracking and is defined only
    in a separate Tcl thread named <code>throttleThread</code>. 
    For each minute within the specified <code>time-window</code> an instance
    of this class exists keeping various statistics. 
    When a minute ends the instance dropping out of the
    time window is destroyed. The procs of this class can be
    used to obtain various kinds of information.

    @author Gustaf Neumann
    @cvs-id $Id: throttle_mod-procs.tcl,v 1.2 2005/07/07 14:09:16 nsadmin Exp $
  }
  Users ad_proc totalInfo {}  {
    Return a list of lists containing information about current 
    users. This list contains these users who have used the server within the 
    monitoring time window (per default: 10 minutes).
    <p>
    For each user a list containing user_id (or peer address), his last
    peer address, the last timestamp, the number of hits, a list
    of values for the activity calculations and the number of ip-switches
    the user is returned. If the user is not logged in, the peer address
    is reported instead of the user_id.
    <p>
    The activity calculations are performed on base of an exponential smoothing
    algorithm which is calculated through an aggregated value, a timestamp 
    (in minutes) and the number of hits in the monitored time window.
    @return list with detailed user info
  } {
    set info [list]
    foreach key [my array names pa] {
      set entry [list $key [my set pa($key)]]
      foreach var [list timestamp hits expSmooth switches] {
	set k ${var}($key)
	lappend entry [expr {[my exists $k] ? [my set $k] : 0}]
      }
      lappend info $entry
    }
    return $info
  }
  Users ad_proc totalUsers {} {
    @return list of request keys (authenticated and unauthenticated users) in time window
  } {
    return [my array names pa]
  }
  Users ad_proc total {} {
    @return number of active users (in time window)
  } {
    return [my array size pa]
  }
  Users ad_proc hits {uid} {
    @param uid request key
    @return Number of hits by this user (in time window)
  } {
    if {[my exists hits($uid)]} {return [my set hits($uid)]} else {return 0}
  }
  Users ad_proc last_pa {uid} {
    @param uid request key
    @return last peer address of the specified users
  } {
    if {[my exists pa($uid)]} { return [my set pa($uid)]} else { return "" }
  }
  Users proc last_click {uid} {
    if {[my exists timestamp($uid)]} {return [my set timestamp($uid)]} else {return 0}
  }
  Users proc last_requests {uid} {
    if {[my exists pa($uid)]} { 
       set urls [list]
       foreach i [Users info instances] {
          if {[$i exists urls($uid)]} {
            foreach u [$i set urls($uid)] { lappend urls $u }
          }
       }
       return [lsort -index 0 $urls]
    } else { return "" }
  }
  Users instproc addKey {key pa url} {
    if {[[self class] exists pa($key)]} {
      # check, if the peer address changed
      if {[string compare [[self class] set pa($key)] $pa]} {
	if {[catch {[self class] incr switches($key)}]} {
	  [self class] set switches($key) 1
	}
	# log the change
	set timestamp [clock format [clock seconds]]
	set f [open $::logdir/switches.log a]
	puts $f "$timestamp -- switch $key from\
 		[[self class] set pa($key)] to $pa $url"
	close $f
      }
    }
    if {[catch {my incr active($key)}]} {
      my set active($key) 1
      [self class] incrRefCount $key $pa
    }
    set entry [list [clock seconds] $url $pa]
    if {[catch {my lappend urls($key) $entry}]} {
      my set urls($key) [list $entry]
    }
    if {[catch {[self class] incr hits($key)}]} {
      [self class] set hits($key) 1
    }
  }
  Users instproc destroy {} {
    foreach key [my array names active] {
      [self class] decrRefCount $key [my set active($key)]
    }
    next
  }
  Users proc expSmooth {ts key} {
    set mins [expr {$ts/60}]
    if {[my exists expSmooth($key)]} {
      foreach {_ aggval lastmins hits} [my set expSmooth($key)] break
      set mindiff [expr {$mins-$lastmins}]
      if {$mindiff == 0} {
	incr hits
	set retval [expr {$aggval*0.3 + $hits*0.7}]
      } else {
	set aggval [expr {$aggval*pow(0.3,$mindiff) + $hits*0.7}]
	set hits 1
      }
    } else {
      set hits 1
      set aggval 1.0
    }
    if {![info exists retval]} {set retval $aggval}
    my set expSmooth($key) [list $retval $aggval $mins $hits]
    return $retval
  }
  Users proc incrRefCount {key pa} {
    if {[my exists refcount($key)]} {
      my incr refcount($key)
    } else {
      my set refcount($key) 1
    }
    my set pa($key) $pa
    my set timestamp($key) [clock seconds]
  }
  Users proc decrRefCount {key hitcount} {
    if {[my exists refcount($key)]} {
      set x [my incr refcount($key) -1]
      my incr hits($key) -$hitcount
      if {$x < 1} {
	my unset refcount($key)
	my unset pa($key)
        catch {my unset expSmooth($key)}
        catch {my unset switches($key)}
      }
    } else {
      my log "+++ cannot decrement refcount for '$key' by $hitcount"
    }
  }
  Users proc perDay {} {
    set ip 0; set auth 0
    foreach i [my array names timestamp] {
      if {[string match *.* $i]} {incr ip} {incr auth}
    }
    return [list $ip $auth]
  }
  
  Users proc perDayCleanup {} {
    set secsPerDay [expr {3600*24}]
    foreach i [lsort [my array names timestamp]] {
      set secs [expr {[clock seconds]-[my set timestamp($i)]}]
      # my log "--- $i: last click $secs secs ago"
      if {$secs>$secsPerDay} {
	foreach {d h m s} [clock format [expr {$secs-$secsPerDay}] \
			       -format {%j %H %M %S}] break
	regexp {^[0]+(.*)$} $d match d
	regexp {^[0]+(.*)$} $h match h
	incr d -1
	incr h -1
	my log "--- $i expired $d days $h hours $m minutes ago"
	my unset timestamp($i)
      }
    }
    after [expr {60000*60}] [list [self] [self proc]]
  }
  Users perDayCleanup
 
  # for debugging purposes: return all running timers
  proc showTimers {} {
     set _ ""
     foreach t [after info] { append _ "$t [after info $t]\n" }
     return $_
  }


  if {[string compare "" [info command ::site_node::get]]} {
    # Figure out the package_id from the site map.
    # There must be a simpler way to do this...
    set id [apm_version_id_from_package_key xotcl-request-monitor]
    set mount_url [site_node::get_children -all \
		       -package_key xotcl-request-monitor -node_id $id]
    array set site_node [site_node::get -url $mount_url]
    set ::package_id $site_node(package_id)
    ns_log notice "+++ package_id of xotcl-request-monitor is $::package_id"
  }

  set logdir [parameter::get -package_id $::package_id \
		  -parameter log-dir \
		  -default [file dirname [file root [ns_config ns/parameters ServerLog]]]]
  if {![file isdirectory $logdir]} {file mkdir $logdir}

} -persistent 1 -ad_doc {
  This is a small request-throttle application that handles simple 
  DOS-attracks on an AOL-server.  A user (request key) is identified 
  via ipAddr or some other key, such as an authenticated userid.
  <p>
  XOTcl Parameters for Class <a 
  href='/xotcl/show-object?object=%3a%3athrottleThread+do+%3a%3aThrottle'>Throttle</a>:
  <ul>
  <li><em>timeWindow:</em>Time window for computing detailed statistics; can 
     be configured via OACS parameter <code>time-window</code></li>
  <li><em>timeoutMs:</em> Time window to keep statistics for a user</li>
  <li><em>startThrottle:</em> If user requests more than this #, he is throttled</li>
  <li><em>toMuch:</em> If user requests more than this #, he is kicked out</li>
  </ul>
  The throttler is defined as a class running in a detached thread.
  It can be subclassed to define e.g. different kinds of throttling policies for
  different kind of request keys. Note that the throttle thread itself
  does not block, only the connection thread blocks if necessary (on throttles).
  <p>
  The controlling thread contains the classes 
    <a href='/xotcl/show-object?object=%3a%3athrottleThread+do+%3a%3aUsers'>Users</a>,
   <a href='/xotcl/show-object?object=%3a%3athrottleThread+do+%3a%3aThrottle'>Throttle</a>,
   <a href='/xotcl/show-object?object=%3a%3athrottleThread+do+%3a%3aCounter'>Counter</a>, 
  <a href='/xotcl/show-object?object=%3a%3athrottleThread+do+%3a%3aMaxCounter'>MaxCounter</a>, ...
  @author Gustaf Neumann
  @cvs-id $Id: throttle_mod-procs.tcl,v 1.2 2005/07/07 14:09:16 nsadmin Exp $
}

::xotcl::THREAD::Client create throttle -server throttleThread -ad_doc {
  This is a client object running in a connection thread. This object
  communicates with the detached server who maintains the various 
  statistics. 

  A call to 'throttle check' can be placed e.g. in the RP or in a
  postauth filter to check throtteling or blocking of requests.
}


if {[info command ::xotcl::THREAD::Client] eq ""} {
  # fallback tcl proc, in case, xotcl or the thread library
  # is not available due to installation problems; "throttle" is 
  # registered as a filter of the request processor. if not available
  # no request can be processed 
  #proc throttle args {
  #  my log "*** fallback procedure of throttle"
  #  my log "*** throttle object not available most probably due to a installation problem of libthread or xotcl"
  #  my log "*** please fix your installation to get the request monitor working"
  #  return filter_ok
  #}
}

throttle proc ms {} {
  expr {[clock clicks -milliseconds] - [my set start_clicks]}
}

throttle proc get_context {} {
  my instvar url requestor user pa query start_clicks

  set url [ad_conn url]
  set query [ad_conn query]
  if {$query ne ""} {
     append url ?$query
  }
  set pa [ad_conn peeraddr]
  set requestor [ad_get_user_id]
  # if user not authorized, use peer address as user id
  if {$requestor == 0} {
    set requestor $pa
    set user "client from $pa"
  } else {
    set user "<a href='/acs-admin/users/one?user_id=$requestor'>$requestor</a>"
  }
  if {$start_clicks eq ""} {
    my log "#### ad-clicks not set ([ad_conn start_clicks]), setting it to clicks='[clock clicks -milliseconds]' <$url> pa=$pa, requestor=<$requestor>"
    set start_clicks [clock clicks -milliseconds]
  }
  #my log "### context url=$url, query=$query pa=$pa requestor=$requestor package=[ad_conn package_id]"
  #show_stack
}

proc show_stack {{m 100}} {
  if {[::info exists ::template::parse_level]} {
    set parse_level $::template::parse_level
  } else {
    set parse_level ""
  }
  set msg "### tid=[::thread::id] <$parse_level> connected=[ns_conn isconnected] "
  if {[ns_conn isconnected]} {
    append msg "flags=[ad_conn flags] status=[ad_conn status] req=[ad_conn request]"
  }
  my log $msg
  set max [info level]  
  if {$m<$max} {set max $m}
  my log "### Call Stack (level: command)"
  for {set i 0} {$i < $max} {incr i} {
    if {[catch {set s [uplevel $i self]} msg]} {
      set s ""
    }
    my log "### [format %5d -$i]:\t$s [info level [expr {-$i}]]"
  }
}

throttle ad_proc check {} {
  This method should be called once per request that is monitored.
  It should be called after authentication shuch we have already 
  the userid if the user is authenticated
} {
  my instvar url requestor user pa query
  my get_context

  foreach {toMuch ms repeat} \
      [my do check $requestor $pa $url \
	   [my set start_clicks] [ns_guesstype $url]] \
      break
  if {$repeat} {
    return -1
  } elseif {$toMuch} {
    my log "*** we have to refuse user $requestor with $toMuch requests"
    my do add_stat reject $user $pa $url $query
    return $toMuch
  } elseif {$ms} {
    my log "*** we have to block user $requestor for $ms ms"
    my do add_stat throttle $user $pa $url $query
    after $ms
    my log "*** continue for user $requestor"
  }
  return 0
}
####
# the following procs are forwarder to the monitoring thread
# for conveniance
####
throttle proc purge_access_stats {} {
  my do purge_access_stats [clock seconds]
}

throttle proc statistics {} {
  my do statistics
}

throttle proc url_statistics {} {
  my do url_statistics
}

####
# the next procs are for the filters (initialized by the -init file)
####
throttle proc preauth args {
  my set start_clicks [ad_conn start_clicks]
  return filter_ok
}

throttle proc postauth args {
  #my log "+++ [self proc] <$args> auth ms [my ms] [ad_conn isconnected]"
  set r [my check]
  if {$r<0} {
    ns_return 200 text/html "
      <H1>Repeated Operation</H1>
      Operation blocked.
      Don't submit the same query, while the old one is still
      running...<p>"
    return filter_return
  } elseif {$r>0} {
    ns_return 200 text/html "
      <H1>Invalid Operation</H1>
      This web server is only open for interactive usage.<br>
      Automated copying and mirroring is not allowed!<p>
      Please slow down your requests...<p>"
    return filter_return
  } else {
    return filter_ok
  }
}
throttle proc trace args {
  #my log "+++ [self proc] <$args>"
  if {[my exists url]} {
    #my log "+++ [self proc] <$args> [my set url] [my ms] [ad_conn isconnected]"
    my do add_url_stat [my set url] [my ms] [my set requestor]
  }
  return filter_ok
}
